import Flutter
import AVFoundation


enum Methods: String {
  case start
  case stop
  case outputFrame
}

enum Channels: String {
  case inputChannelName = "com.wearemobilefirst.audio_io.inputAudio";
  case outputChannelName = "com.wearemobilefirst.audio_io.outputAudio";
  case methodChannelName = "com.wearemobilefirst.audio_io";
}

public class SwiftAudioIoPlugin: NSObject, FlutterPlugin, FlutterStreamHandler {

    let engine = AVAudioEngine()
    var converter = AVAudioMixerNode()

    var _inputEventChannel: FlutterEventChannel?
    var _outputEventChannel: FlutterEventChannel? 
    var _inputEventSink: FlutterEventSink?
    var buffer = RingBuffer<Float>(count: 48000);
    let queue = DispatchQueue(label: "SwiftAudioIoPluginQueue", attributes: .concurrent)

    private lazy var sourceNode = AVAudioSourceNode { (_, _, frameCount, audioBufferList) -> OSStatus in
        let ablPointer = UnsafeMutableAudioBufferListPointer(audioBufferList)
        self.queue.sync() {
        for frame in 0..<Int(frameCount) {
            for buffer in ablPointer {
                let buf: UnsafeMutableBufferPointer<Float> = UnsafeMutableBufferPointer(buffer)
                buf[frame] = self.buffer.read() ?? 0
            }
        }
        }
        return noErr
    }
    
    private lazy var sinkNode = AVAudioSinkNode() { (timeStamp, frames, audioBufferList) ->
           OSStatus in
             let ptr = audioBufferList.pointee.mBuffers.mData?.assumingMemoryBound(to: Float.self)
             var monoSamples = Array(UnsafeBufferPointer(start: ptr, count: Int(frames)))
             var doubleSamples = [Double](repeating: 0.0, count: monoSamples.count)
              for frame in 0..<Int(monoSamples.count) {
                doubleSamples[frame] = Double(monoSamples[frame]);
              }
              self._inputEventSink?(doubleSamples)
             return noErr
         }

  public static func register(with registrar: FlutterPluginRegistrar) {
    let channel = FlutterMethodChannel(name: Channels.methodChannelName.rawValue, binaryMessenger: registrar.messenger())
    let instance = SwiftAudioIoPlugin()
    registrar.addMethodCallDelegate(instance, channel: channel)
    instance._inputEventChannel = FlutterEventChannel(name: Channels.inputChannelName.rawValue , binaryMessenger: registrar.messenger())
    instance._inputEventChannel?.setStreamHandler(instance)                                                                 
    instance._outputEventChannel = FlutterEventChannel(name: Channels.outputChannelName.rawValue , binaryMessenger: registrar.messenger())        
                                                      
  }

  public func onListen(withArguments arguments: Any?,
                       eventSink: @escaping FlutterEventSink) -> FlutterError? {
        _inputEventSink = eventSink; 
    return nil            
  }

  public func onCancel(withArguments arguments: Any?) -> FlutterError? {
    _inputEventSink = nil
    return nil
  }

  public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
    switch call.method {
    case Methods.start.rawValue :
      self.start()
      result(nil)
      break;
    case Methods.stop.rawValue:
      self.stop()
      result(nil)
      break;
    case Methods.outputFrame.rawValue:
      guard let data = call.arguments as? Array<Double> else{
        result(nil)
        break; 
      };
      queue.async(flags: .barrier) {
        for frame in 0..<data.count {
          _ = self.buffer.write(Float(data[frame]))
        }
      }
      result(nil)
      break; 
    default:
      result(FlutterMethodNotImplemented)
    }

  }
                                                                                                                                                                                      
  public func stop() {
     engine.stop()
  }

  public func start() {
    // Setup AVAudioSession
        do {
            try AVAudioSession.sharedInstance().setCategory(AVAudioSession.Category.playAndRecord)

            let ioBufferDuration = 128.0 / 44100.0

            try AVAudioSession.sharedInstance().setPreferredIOBufferDuration(ioBufferDuration)

        } catch {
            assertionFailure("AVAudioSession setup error: \(error)")
        }
        converter.outputVolume = 1.0
        // Setup engine and node instances
        let input = engine.inputNode
        let output = engine.mainMixerNode
        let inputFormat = input.outputFormat(forBus: 0)
        let outputFormat = output.inputFormat(forBus: 0)
        print(inputFormat)
        print(outputFormat)

        // Connect nodes
        engine.attach(converter)
        engine.attach(sinkNode)
        engine.attach(sourceNode)
        engine.connect(input, to: sinkNode, format: nil)
        engine.connect(sourceNode, to: converter, format: inputFormat)
        engine.connect(converter, to: output, format: outputFormat)

        // Start engine
        do {
            try engine.start()
        } catch {
            assertionFailure("AVAudioEngine start error: \(error)")
        }
  }
}



public struct RingBuffer<T> {
  fileprivate var array: [T?]
  fileprivate var readIndex = 0
  fileprivate var writeIndex = 0

  public init(count: Int) {
    array = [T?](repeating: nil, count: count)
  }

  public mutating func write(_ element: T) -> Bool {
    if !isFull {
      array[writeIndex % array.count] = element
      writeIndex += 1
      return true
    } else {
      return false
    }
  }

  public mutating func read() -> T? {
    if !isEmpty {
      let element = array[readIndex % array.count]
      readIndex += 1
      return element
    } else {
      return nil
    }
  }

  fileprivate var availableSpaceForReading: Int {
    return writeIndex - readIndex
  }

  public var isEmpty: Bool {
    return availableSpaceForReading == 0
  }

  fileprivate var availableSpaceForWriting: Int {
    return array.count - availableSpaceForReading
  }

  public var isFull: Bool {
    return availableSpaceForWriting == 0
  }
}


